# grab_sotus.py
# Lillian Lee, LJL2, Mar 2022
#-*- coding: utf-8 -*-  # for "unusual" characters


"""Get real U.S. State of the Union (SOTU) data for CS1110 students to try out
    their A3 code on!  Puts into file sotu_dict.py a dictionary with keys=years,
    values=tokenlists of corresponding SOTU addresses.

    Data comes from the American Presidency Project. They point out that 
        "Recent presidents starting with Reagan, addressed a joint session of Congress shortly after their inaugurations and these messages are, technically, not considered to be "State of the Union" addresses."
    We include these "asterisk" messages in our data.

    We omit years where there was more than one version.

    There will be some stray text, such as interruptions by the audience, that we
    do not filter out.

"""


import requests # for getting webpages
import string # for removing punctuation
import pprint # for nice formatting on the dictionary

# For getting the URLs for pages with individual SOTU addresses
# Inferred by looking at html from the American Presidency Project
SOTUS_PG_LINES = requests.get("https://www.presidency.ucsb.edu/documents/presidential-documents-archive-guidebook/annual-messages-congress-the-state-the-union#Table%20of%20SOTU").text.split("\n")
YEAR_LINE_STARTER = '<td style="text-align: center; height: 16px;"><a href="'
YEAR_LINE_ENDER = '</a></td>'
YEAR_LINE_ENDER2 = '</a>*</td>' # Some years are asterisked

# Years w/ multiple SOTUs
YEAR_LINE_IGNORES = ["1981",
                     "1980",
                     "1979",
                     "1978",
                     "1974",
                     "nixon1973", 
                     "1972",
                     "1956",
                     "1945",
                     "1790"]

def get_tokenlist(url):
    """Returns tokenlist of speech text given at `url`.

    Precondition: `url` is the URL for a page containing SOTU text between
    the lines <div class="field-docs-content"> and </div>, possibly with some
    <p>, </p>, <i>, and </i> markup.
    """

    m1='<div class="field-docs-content">'
    m2 = '</div>'
    text = requests.get(url).text
    i1 = text.index(m1)+len(m1)
    i2 = text.index(m2, i1)
    return _listit(text[i1:i2])


def get_sotu_tokenlists():
    """Returns dictionary with keys= strings that are year labels, values =
       tokenlist for the U.S. Presidential State of the Union (SOTU) address 
       that year.
       The dictionary is in order of increasing year, starting from 1982
       (1981 had both a written and a spoken one)"""


    url_list = []  # will be SOTU_url objects, sorted by increasing year

    for line in SOTUS_PG_LINES:
        line = line.strip()

        # One would ordinarily use "regular expressions" for this task, but
        # we don't learn about them in CS1110. The code below is meant to
        # be accessible to CS1110 students.

        if line.startswith(YEAR_LINE_STARTER) and not _ignorable_line(line):
            if (line.endswith(YEAR_LINE_ENDER)):
                line = _scoop_yline(line, 1)
            elif  line.endswith(YEAR_LINE_ENDER2):
                line = _scoop_yline(line, 2)
            [url, year] = line.split('">') # year is a string
            url_list.append(SOTU_url(int(year), url))

        url_list.sort()

    out_dict = {}
    for url_obj in url_list:
        y = url_obj.year
        print('working on year ' + str(y))
        out_dict[y] = get_tokenlist(url_obj.url)
        #@ print('first 100 words of that year: ' + str(out_dict[y][:100]))

    return out_dict


# Objects representing URLs that can be sorted by year
class SOTU_url:
    """Attributes: 
        year: int
        url: string (source URL)"""

    def __init__(self, y, u):
        """New SOTU_url w/ year y and url u"""
        assert type(y) == int
        self.year = y 
        self.url = u

    def __lt__(self, other):
        if not isinstance(other, SOTU_url):
            return NotImplemented
        else:
            return self.year < other.year     

    def __repr__(self):
        return "SOTU_URL("+str(self.year)+", "+self.url+")"




############ Helpers ##############
def _scoop_yline(line, etype):
    """If type is 1,
        Return part of line between YEAR_LINE_STARTER and YEAR_LINE_ENDER.
       If type is 2,
        Return part of line between YEAR_LINE_STARTER and YEAR_LINE_ENDER2.

    Precondition: 
    `line` starts with YEAR_LINE_STARTER, ends with YEAR_LINE_ENDER or 
     YEAR_LINE_ENDER2 as specified by `etype`, which must be either int 1 or
     int 2. """
    assert YEAR_LINE_STARTER in line 

    i1 = len(YEAR_LINE_STARTER)
    if etype==1:
        i2 = line.rindex(YEAR_LINE_ENDER)
    else:
        assert etype ==2
        i2 = line.rindex(YEAR_LINE_ENDER2)
    return line[i1:i2]


def _ignorable_line(line):
    """Returns True if line contains item in YEAR_LINE_IGNORES, False o.w."""
    for excludable in YEAR_LINE_IGNORES:
        if excludable in line:
            return True
    # if got here, didn't hit an excludable
    return False




    # get the wordlists for each year, plus the labsl

    # would be nice to output a dictionary indexed by years.  Would also 
    # be nice to know the first and last year in the dictionary.
    # Would also be nice for the dictionary to be sorted by year

# Because of interruptions/stutters, need to remove hyphens
def _listit(s):
    """lowercase, remove text between <i> and </i>, remove <p> and </p>,
    strip punctuation except apostrophes, 
    strip numbers, return a list of tokens in string `s`
    """
    s = _remove_italics(s.lower()) 

    # replace special hyphens with space (handling interruptions/repeats)
    s = s.replace('—',' ')

    # Get rid of html tags
    # (Pray that there aren't any ">" or "<" in SOTU speeches)
    inside_html = False
    
    temp = "" # building copy of s that doesn't have <...> in it
    for c in s: 
        if c == '<':
            inside_html=True
        elif c == '>':
            inside_html = False
        elif inside_html:
            pass
        else:
            assert inside_html == False
            temp = temp + c

    # replace punctuation (except apostrophe) and digits with space
    delete_chars = string.punctuation.replace("'", '')
    delete_chars += string.digits
    deletemap=''.maketrans(delete_chars, " "*len(delete_chars))
    return temp.translate(deletemap).split()



def _remove_italics(s):
    """return version of `s` in which items between italics are removed"""
    if '<i>' in s:
        still_italics = True
    else:
        still_italics = False

    out = ''
    search = s #parts of s to still check/extract from
    while still_italics and len(search) > 0:
        next_italics_pos = search.index('<i>')
        out = out + search[:next_italics_pos]
        after_next_italics_pos = search.index('</i>')+len('</i>')
        search = search[after_next_italics_pos:]
        if '<i>' not in search:
            still_italics = False
    return out + search # don't forget the remaining stuff, which has no <i>s

if __name__ == '__main__':
    print('\n\n===================')
    d = get_sotu_tokenlists()
    with open('sotu_dict.py', 'w') as out:

        out.write('# sotu_dict.py\n')
        out.write('# Prof. Lee, LJL2, Mar 2022\n\n')
        out.write('"""dictionary of (most) SOTU addresses\n\n\n')
        out.write('Original data source: https://www.presidency.ucsb.edu/documents/presidential-documents-archive-guidebook/annual-messages-congress-the-state-the-union\n\n\n')
        out.write('Processing code: grab_sotus.py\n"""\n')
        out.write('SOTU_DICT = \\\n')

        pprint.pprint(d, stream=out)
